# -*- coding: UTF-8 -*-

import sys

from typing import Optional, Literal, Union, Tuple, Dict, List, Any

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from typing import Literal
else:  # pragma: no cover (py38+)
    from typing_extensions import Literal

import re
import json
from loguru import logger
from curl_cffi import requests

from ..magneto.response import Response
from ..magneto.session import Session, AsyncSession


def crack_shape_v1(
    user_token: str, href: str, proxy: Optional[str] = None,
    script_url: Optional[str] = None, script_regexp: Optional[str] = None, 
    vmp_url: Optional[str] = None, vmp_regexp: Optional[str] = None, fast: bool = False,
    internal_host: bool = True, branch: Optional[str] = None,
    developer_id: Optional[str] = None, auth: bool = False,
    session: Optional[Session] = None, cookies: Dict[str, str] = {},
    timeout: int = 15, allow_redirects: bool = True, verify: bool = False, http_version: Optional[Literal["http1", "http2"]] = None,
    user_agent: Union[str, Tuple[int, Literal["windows", "macos"], Literal["chrome", "edge"]]] = None,
    ipinfo: Dict[str, str] = {}, extra: Dict[str, str] = {}, debug: bool = False
) -> Tuple[Session, Dict[str, str], Dict[str, Any]]:   
    if not session:
        session = Session(
            user_agent=user_agent, 
            proxy=proxy, 
            ipinfo=ipinfo, 
            cookies=cookies, 
            timeout=timeout, 
            allow_redirects=allow_redirects, 
            verify=verify, 
            http_version=http_version
        )
        if not ipinfo:
            session.update_ipinfo()

    if not extra:
        extra.update(session.client_hints)
        
    user_agent = session.user_agent
    proxy = session.proxy
    
    domain = href.split("/")[2]
    nocaptcha_host = "api.nocaptcha.cn" if internal_host else "api.nocaptcha.io"
    if not script_url:
        data = {
            "method": "read",
            "key": domain,
        }
        site_arg = requests.post(
            f"http://{nocaptcha_host}/api/wanda/shape/p",
            json=data
        ).text
        
        if site_arg:
            site_arg = json.loads(site_arg)
            script_url = site_arg.get("script_url")
            vmp_url = site_arg.get("vmp_url")
            vmp_regexp = site_arg.get("vmp_regexp")
    
    origin = "/".join(href.split("/")[0:3])

    if not script_url:
        headers = [
            f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
            "sec-ch-ua-mobile: ?0",
            f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
            "upgrade-insecure-requests: 1",
            f"user-agent: {user_agent}",
            "accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
            "sec-fetch-site: none",
            "sec-fetch-mode: navigate",
            "sec-fetch-user: ?1",
            "sec-fetch-dest: document",
            "accept-encoding: gzip, deflate, br, zstd",
            f"accept-language: {session.client_hints['accept-language']}",
            "priority: u=0, i"
        ]
        if script_regexp:
            headers = [
                f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
                f"user-agent: {user_agent}",
                f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
                "sec-ch-ua-mobile: ?0",
                "accept: */*",
                "sec-fetch-site: same-origin",
                "sec-fetch-mode: no-cors",
                "sec-fetch-dest: script",
                f"referer: {href}",
                "accept-encoding: gzip, deflate, br, zstd",
                f"accept-language: {session.client_hints['accept-language']}",
                "priority: u=1"
            ]
            html = session.get(href, headers=headers).text
            script_url = re.findall(script_regexp, html)[1]
                
    if not script_url.startswith('http'):
        script_url = origin + script_url
    
    try:
        headers = [
            f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
            f"user-agent: {user_agent}",
            f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
            "sec-ch-ua-mobile: ?0",
            "accept: */*",
            "sec-fetch-site: same-origin",
            "sec-fetch-mode: no-cors",
            "sec-fetch-dest: script",
            f"referer: {href}",
            "accept-encoding: gzip, deflate, br, zstd",
            f"accept-language: {session.client_hints['accept-language']}",
            "priority: u=1"
        ]
        resp = session.get(script_url, headers=headers)
        if resp.status_code != 200:
            raise Warning(f"初始脚本状态码异常: {resp.status_code}")
        script = resp.text
        if script == "fail":
            raise Warning("初始脚本获取异常")
    except Warning as e:
        raise e
    except:
        raise Warning("初始脚本获取失败")
    
    if not vmp_url:
        if vmp_regexp:
            try:
                vmp_url = re.search(vmp_regexp, script)[1]
            except:
                raise Warning('vmp 地址获取失败')
    
    vmp_script = None
    if vmp_url:
        if not vmp_url.startswith("http"):
            vmp_url = origin + vmp_url
        try:
            vmp_resp = session.get(vmp_url, headers=headers)
            if vmp_resp.status_code != 200:
                raise Warning("vmp 脚本请求失败")
            
            vmp_script = vmp_resp.text
        except:
            raise Warning("vmp 获取失败")
    
    nocaptcha_resp = requests.post(
        f'http://{nocaptcha_host}/api/wanda/shape/v1', headers={
            'user-token': user_token,
            **({ "Developer-Id": developer_id } if developer_id else {})
        }, json={
            'is_auth': auth,
            "branch": branch,
            "href": href,
            "script_url": script_url,
            "script_content": script,
            "vmp_url": vmp_url,
            "vmp_content": vmp_script,
            "user_agent": user_agent,
            **session.ipinfo,
            "fast": fast,
            "cookies": session.cookies,
        }
    ).json()
    if debug:
        logger.debug(nocaptcha_resp)
    if nocaptcha_resp["status"]:
        extra.update(nocaptcha_resp["extra"])
        shape_headers = nocaptcha_resp["data"]
        return session, shape_headers, extra
    else:
        raise Warning(f'计算失败, id: {nocaptcha_resp["id"]}, err: {nocaptcha_resp["msg"]}')


async def async_crack_shape_v1(
    user_token: str, href: str, proxy: Optional[str] = None,
    script_url: Optional[str] = None, script_regexp: Optional[str] = None,
    vmp_url: Optional[str] = None, vmp_regexp: Optional[str] = None, fast: bool = False,
    internal_host: bool = True, branch: Optional[str] = None,
    developer_id: Optional[str] = None, auth: bool = False,
    session: Optional[AsyncSession] = None, cookies: Dict[str, str] = {},
    timeout: int = 15, allow_redirects: bool = True, verify: bool = False, http_version: Optional[Literal["http1", "http2"]] = None,
    user_agent: Union[str, Tuple[int, Literal["windows", "macos"], Literal["chrome", "edge"]]] = None,
    ipinfo: Dict[str, str] = {}, extra: Dict[str, str] = {}, debug: bool = False
) -> Tuple[AsyncSession, Dict[str, str], Dict[str, Any]]:    
    if not session:
        session = AsyncSession(
            user_agent=user_agent, 
            proxy=proxy, 
            ipinfo=ipinfo, 
            cookies=cookies, 
            timeout=timeout, 
            allow_redirects=allow_redirects, 
            verify=verify, 
            http_version=http_version
        )
        if not ipinfo:
            await session.update_ipinfo()

    if not extra:
        extra.update(session.client_hints)
        
    user_agent = session.user_agent
    proxy = session.proxy
    
    nocaptcha_host = "api.nocaptcha.cn" if internal_host else "api.nocaptcha.io"
    domain = href.split("/")[2]
    async with requests.AsyncSession() as _session:
        if not script_url:
            data = {
                "method": "read",
                "key": domain,
            }
            site_arg = (await _session.post(
                f"http://{nocaptcha_host}/api/wanda/shape/p",
                json=data
            )).text
            
            if site_arg:
                site_arg = json.loads(site_arg)
                script_url = site_arg.get("script_url")
                vmp_url = site_arg.get("vmp_url")
                vmp_regexp = site_arg.get("vmp_regexp")
        
        origin = "/".join(href.split("/")[0:3])

        if not script_url:
            headers = [
                f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
                "sec-ch-ua-mobile: ?0",
                f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
                "upgrade-insecure-requests: 1",
                f"user-agent: {user_agent}",
                "accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
                "sec-fetch-site: none",
                "sec-fetch-mode: navigate",
                "sec-fetch-user: ?1",
                "sec-fetch-dest: document",
                "accept-encoding: gzip, deflate, br, zstd",
                f"accept-language: {session.client_hints['accept-language']}",
                "priority: u=0, i"
            ]
            if script_regexp:
                headers = [
                    f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
                    f"user-agent: {user_agent}",
                    f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
                    "sec-ch-ua-mobile: ?0",
                    "accept: */*",
                    "sec-fetch-site: same-origin",
                    "sec-fetch-mode: no-cors",
                    "sec-fetch-dest: script",
                    f"referer: {href}",
                    "accept-encoding: gzip, deflate, br, zstd",
                    f"accept-language: {session.client_hints['accept-language']}",
                    "priority: u=1"
                ]
                resp = await session.get(href, headers=headers)
                script_url = re.findall(script_regexp, await resp.async_text())[1]
                  
        if not script_url.startswith('http'):
            script_url = origin + script_url
        
        try:
            headers = [
                f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
                f"user-agent: {user_agent}",
                f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
                "sec-ch-ua-mobile: ?0",
                "accept: */*",
                "sec-fetch-site: same-origin",
                "sec-fetch-mode: no-cors",
                "sec-fetch-dest: script",
                f"referer: {href}",
                "accept-encoding: gzip, deflate, br, zstd",
                f"accept-language: {session.client_hints['accept-language']}",
                "priority: u=1"
            ]
            resp = await session.get(script_url, headers=headers)
            if resp.status_code != 200:
                raise Warning(f"初始脚本状态码异常: {resp.status_code}")
            script = await resp.async_text()
            if script == "fail":
                raise Warning("初始脚本获取异常")
        except Warning as e:
            raise e
        except:
            raise Warning("初始脚本获取失败")
        
        if not vmp_url:
            if vmp_regexp:
                try:
                    vmp_url = re.search(vmp_regexp, script)[1]
                except:
                    raise Warning('vmp 地址获取失败')
        
        vmp_script = None
        if vmp_url:
            if not vmp_url.startswith("http"):
                vmp_url = origin + vmp_url
            try:
                vmp_resp = await session.get(vmp_url, headers=headers)
                if vmp_resp.status_code != 200:
                    raise Warning("vmp 脚本请求失败")
                
                vmp_script = await vmp_resp.async_text()
            except:
                raise Warning("vmp 获取失败")
        
        nocaptcha_resp = (await _session.post(
            f'http://{nocaptcha_host}/api/wanda/shape/v1', headers={
                'user-token': user_token,
                **({ "Developer-Id": developer_id } if developer_id else {})
            }, json={
                'is_auth': auth,
                "branch": branch,
                "href": href,
                "script_url": script_url,
                "script_content": script,
                "vmp_url": vmp_url,
                "vmp_content": vmp_script,
                "user_agent": user_agent,
                **session.ipinfo,
                "fast": fast,
                "cookies": session.cookies,
            }
        )).json()
        if debug:
            logger.debug(nocaptcha_resp)
        if nocaptcha_resp["status"]:
            extra.update(nocaptcha_resp["extra"])
            shape_headers = nocaptcha_resp["data"]
            return session, shape_headers, extra
        else:
            raise Warning(f'计算失败, id: {nocaptcha_resp["id"]}, err: {nocaptcha_resp["msg"]}')


def crack_shape_v2(
    user_token: str, href: str, pkey: Optional[str] = None, proxy: Optional[str] = None, request: Optional[Dict[str, str]] = None,
    script_url: Optional[str] = None, script_regexp: Optional[str] = None, vmp_url: Optional[str] = None, vmp_regexp: Optional[str] = None, 
    fast: bool = False, action: Optional[str] = None,
    internal_host: bool = True, branch: Optional[str] = None,
    developer_id: Optional[str] = None, auth: bool = False,
    session: Optional[Session] = None, cookies: Dict[str, str] = {},
    timeout: int = 15, allow_redirects: bool = True, verify: bool = False, http_version: Optional[Literal["http1", "http2"]] = None,
    user_agent: Union[str, Tuple[int, Literal["windows", "macos"], Literal["chrome", "edge"]]] = None,
    ipinfo: Dict[str, str] = {}, extra: Dict[str, str] = {}, debug: bool = False
) -> Tuple[Session, Union[Dict[str, str], Response], Dict[str, Any]]: 
    if not session:
        session = Session(
            user_agent=user_agent, 
            proxy=proxy, 
            ipinfo=ipinfo, 
            cookies=cookies, 
            timeout=timeout, 
            allow_redirects=allow_redirects, 
            verify=verify, 
            http_version=http_version
        )
        if not ipinfo:
            session.update_ipinfo()

    if not extra:
        extra.update(session.client_hints)
        
    user_agent = session.user_agent
    proxy = session.proxy
    
    nocaptcha_host = "api.nocaptcha.cn" if internal_host else "api.nocaptcha.io"

    origin = "/".join(href.split("/")[0:3])

    headers = [
        f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
        "sec-ch-ua-mobile: ?0",
        f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
        "upgrade-insecure-requests: 1",
        f"user-agent: {user_agent}",
        "accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
        "sec-fetch-site: none",
        "sec-fetch-mode: navigate",
        "sec-fetch-user: ?1",
        "sec-fetch-dest: document",
        "accept-encoding: gzip, deflate, br, zstd",
        f"accept-language: {session.client_hints['accept-language']}",
        "priority: u=0, i"
    ]
    resp = session.get(href, headers=headers)
    html = resp.text
    
    if pkey:
        if not script_url:
            data = {
                "method": "read",
                "key": pkey.lower(),
            }
            site_arg = requests.post(
                f"http://{nocaptcha_host}/api/wanda/shape/p",
                json=data
            ).text
            
            if site_arg:
                site_arg = json.loads(site_arg)
                script_url = site_arg.get("script_url")
                vmp_url = site_arg.get("vmp_url")
                vmp_regexp = site_arg.get("vmp_regexp")
                if not request:
                    request = site_arg.get("request")
        
        if script_regexp:
            script_url = re.findall(script_regexp, html)[1]
                    
        if not script_url.startswith('http'):
            script_url = origin + script_url
        
        try:
            headers = [
                f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
                f"user-agent: {user_agent}",
                f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
                "sec-ch-ua-mobile: ?0",
                "accept: */*",
                "sec-fetch-site: same-origin",
                "sec-fetch-mode: no-cors",
                "sec-fetch-dest: script",
                f"referer: {href}",
                "accept-encoding: gzip, deflate, br, zstd",
                f"accept-language: {session.client_hints['accept-language']}",
                "priority: u=1"
            ]
            resp = session.get(script_url, headers=headers)
            if resp.status_code != 200:
                raise Warning(f"初始脚本状态码异常: {resp.status_code}")
            script = resp.text
            if script == "fail":
                raise Warning("初始脚本获取异常")
        except Warning as e:
            raise e
        except:
            raise Warning("初始脚本获取失败")
        
        if not vmp_url:
            if vmp_regexp:
                try:
                    vmp_url = re.search(vmp_regexp, script)[1]
                except:
                    raise Warning('vmp 地址获取失败')
        
        vmp_script = None
        if vmp_url:
            if not vmp_url.startswith("http"):
                vmp_url = origin + vmp_url
            try:
                vmp_resp = session.get(vmp_url, headers=headers)
                if vmp_resp.status_code != 200:
                    raise Warning("vmp 脚本请求失败")
                
                vmp_script = vmp_resp.text
            except:
                raise Warning("vmp 获取失败")
        
        nocaptcha_resp = requests.post(
            f'http://{nocaptcha_host}/api/wanda/shape/v2', headers={
                'user-token': user_token,
                **({ "Developer-Id": developer_id } if developer_id else {})
            }, json={
                'is_auth': auth,
                "branch": branch,
                "href": href,
                "pkey": pkey,
                "request": request,
                "script_url": script_url,
                "script_content": script,
                "vmp_url": vmp_url,
                "vmp_content": vmp_script,
                "user_agent": user_agent,
                **session.ipinfo,
                "fast": fast,
                "cookies": session.cookies,
                "action": action,
            }
        ).json()
    
    else:
        if 'ISTL-REDIRECT-TO' in html:
            nocaptcha_resp = requests.post(
                f'http://{nocaptcha_host}/api/wanda/shape/v2', headers={
                    'user-token': user_token,
                    **({ "Developer-Id": developer_id } if developer_id else {})
                }, json={
                    'is_auth': auth,
                    "branch": branch,
                    "href": href,
                    "html": html,
                    "user_agent": user_agent,
                    **session.ipinfo,
                    "fast": fast,
                    "cookies": session.cookies,
                }
            ).json()
        else:
            return session, resp, extra
        
    if debug:
        logger.debug(nocaptcha_resp)

    if nocaptcha_resp["status"]:
        extra.update(nocaptcha_resp["extra"])
        if isinstance(nocaptcha_resp["data"], list):
            shape_headers = nocaptcha_resp["data"][0]
        else:
            shape_headers = nocaptcha_resp["data"]['headers']
            
        return session, shape_headers, extra
    else:
        raise Warning(f'计算失败, id: {nocaptcha_resp["id"]}, err: {nocaptcha_resp["msg"]}')


async def async_crack_shape_v2(
    user_token: str, href: str, pkey: Optional[str] = None, proxy: Optional[str] = None, request: Optional[Dict[str, str]] = None,
    script_url: Optional[str] = None, script_regexp: Optional[str] = None, vmp_url: Optional[str] = None, vmp_regexp: Optional[str] = None, 
    fast: bool = False, action: Optional[str] = None,
    internal_host: bool = True, branch: Optional[str] = None,
    developer_id: Optional[str] = None, auth: bool = False,
    session: Optional[AsyncSession] = None, cookies: Dict[str, str] = {},
    timeout: int = 15, allow_redirects: bool = True, verify: bool = False, http_version: Optional[Literal["http1", "http2"]] = None,
    user_agent: Union[str, Tuple[int, Literal["windows", "macos"], Literal["chrome", "edge"]]] = None,
    ipinfo: Dict[str, str] = {}, extra: Dict[str, str] = {}, debug: bool = False
) -> Tuple[AsyncSession, Union[Dict[str, str], Response], Dict[str, Any]]:    
    if not session:
        session = AsyncSession(
            user_agent=user_agent, 
            proxy=proxy, 
            ipinfo=ipinfo, 
            cookies=cookies, 
            timeout=timeout, 
            allow_redirects=allow_redirects, 
            verify=verify, 
            http_version=http_version
        )
        if not ipinfo:
            await session.update_ipinfo()

    if not extra:
        extra.update(session.client_hints)
        
    user_agent = session.user_agent
    proxy = session.proxy
    
    nocaptcha_host = "api.nocaptcha.cn" if internal_host else "api.nocaptcha.io"
    async with requests.AsyncSession() as _session:        
        origin = "/".join(href.split("/")[0:3])

        headers = [
            f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
            "sec-ch-ua-mobile: ?0",
            f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
            "upgrade-insecure-requests: 1",
            f"user-agent: {user_agent}",
            "accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
            "sec-fetch-site: none",
            "sec-fetch-mode: navigate",
            "sec-fetch-user: ?1",
            "sec-fetch-dest: document",
            "accept-encoding: gzip, deflate, br, zstd",
            f"accept-language: {session.client_hints['accept-language']}",
            "priority: u=0, i"
        ]
        resp = await session.get(href, headers=headers)
        html = await resp.async_text()
        
        if pkey:
            if not script_url:
                data = {
                    "method": "read",
                    "key": pkey.lower(),
                }
                site_arg = (await _session.post(
                    f"http://{nocaptcha_host}/api/wanda/shape/p",
                    json=data
                )).text
                
                if site_arg:
                    site_arg = json.loads(site_arg)
                    script_url = site_arg.get("script_url")
                    vmp_url = site_arg.get("vmp_url")
                    vmp_regexp = site_arg.get("vmp_regexp")
                    if not request:
                        request = site_arg.get("request")
            if script_regexp:
                script_url = re.findall(script_regexp, html)[1]
                
            if not script_url.startswith('http'):
                script_url = origin + script_url
            
            try:
                headers = [
                    f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
                    f"user-agent: {user_agent}",
                    f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
                    "sec-ch-ua-mobile: ?0",
                    "accept: */*",
                    "sec-fetch-site: same-origin",
                    "sec-fetch-mode: no-cors",
                    "sec-fetch-dest: script",
                    f"referer: {href}",
                    "accept-encoding: gzip, deflate, br, zstd",
                    f"accept-language: {session.client_hints['accept-language']}",
                    "priority: u=1"
                ]
                resp = await session.get(script_url, headers=headers)
                if resp.status_code != 200:
                    raise Warning(f"初始脚本状态码异常: {resp.status_code}")
                script = await resp.async_text()
                if script == "fail":
                    raise Warning("初始脚本获取异常")
            except Warning as e:
                raise e
            except:
                raise Warning("初始脚本获取失败")
            
            if not vmp_url:
                if vmp_regexp:
                    try:
                        vmp_url = re.search(vmp_regexp, script)[1]
                    except:
                        raise Warning('vmp 地址获取失败')
            
            vmp_script = None
            if vmp_url:
                if not vmp_url.startswith("http"):
                    vmp_url = origin + vmp_url
                try:
                    vmp_resp = await session.get(vmp_url, headers=headers)
                    if vmp_resp.status_code != 200:
                        raise Warning("vmp 脚本请求失败")
                    
                    vmp_script = await vmp_resp.async_text()
                except:
                    raise Warning("vmp 获取失败")
            
            nocaptcha_resp = (await _session.post(
                f'http://{nocaptcha_host}/api/wanda/shape/v2', headers={
                    'user-token': user_token,
                    **({ "Developer-Id": developer_id } if developer_id else {})
                }, json={
                    'is_auth': auth,
                    "branch": branch,
                    "href": href,
                    "pkey": pkey,
                    "request": request,
                    "script_url": script_url,
                    "script_content": script,
                    "vmp_url": vmp_url,
                    "vmp_content": vmp_script,
                    "user_agent": user_agent,
                    **session.ipinfo,
                    "fast": fast,
                    "cookies": session.cookies,
                    "action": action,
                }
            )).json()
        else:
            if 'ISTL-REDIRECT-TO' in html:
                nocaptcha_resp = (await _session.post(
                    f'http://{nocaptcha_host}/api/wanda/shape/v2', headers={
                        'user-token': user_token,
                        **({ "Developer-Id": developer_id } if developer_id else {})
                    }, json={
                        'is_auth': auth,
                        "branch": branch,
                        "href": href,
                        "html": html,
                        "user_agent": user_agent,
                        **session.ipinfo,
                        "fast": fast,
                        "cookies": session.cookies,
                    }
                )).json()
            else:
                return session, resp, extra
        
        if debug:
            logger.debug(nocaptcha_resp)
        if nocaptcha_resp["status"]:
            extra.update(nocaptcha_resp["extra"])
            if isinstance(nocaptcha_resp["data"], list):
                shape_headers = nocaptcha_resp["data"][0]
            else:
                shape_headers = nocaptcha_resp["data"]['headers']
            return session, shape_headers, extra
        else:
            raise Warning(f'计算失败, id: {nocaptcha_resp["id"]}, err: {nocaptcha_resp["msg"]}')



# -*- coding: UTF-8 -*-


from .crackers.akamai import AkamaiV2Cracker, crack_akamai_v3, async_crack_akamai_v3
from .crackers.aws import AwsUniversalCracker
from .crackers.cloudflare import CloudFlareCracker
from .crackers.discord import DiscordCracker
from .crackers.hcaptcha import HcaptchaCracker
from .crackers.incapsula import (IncapsulaRbzidCracker,
                                 IncapsulaReese84Cracker,
                                 IncapsulaUtmvcCracker)
from .crackers.recaptcha import (ReCaptchaAppCracker,
                                 ReCaptchaEnterpriseCracker,
                                 ReCaptchaSteamCracker,
                                 ReCaptchaUniversalCracker)
from .crackers.datadome import crack_datadome, async_crack_datadome
from .crackers.kasada import KasadaCdCracker, crack_kasada, async_crack_kasada
from .crackers.perimeterx import PerimeterxCracker, crack_perimeterx, async_crack_perimeterx
from .crackers.shape import crack_shape_v1, async_crack_shape_v1, crack_shape_v2, async_crack_shape_v2
from .crackers.tls import TlsV1Cracker

from .crackers.unlocker import Unlocker, AsyncUnlocker, run_unlocker, async_run_unlocker
from .crackers.unlocker import KohlCheckBalanceParams, LululemonCheckBalanceParams, WbiParams, DicksCheckBalanceParams
from .crackers.unlocker import FinishlineCheckBalanceParams, FootlockerCheckBalanceParams, ReiCheckBalanceParams
from .crackers.unlocker import One4AllCheckBalanceParams, SaksCheckBalanceParams, CoachCheckBalanceParams
from .crackers.unlocker import FandangoCheckBalanceParams, ColumbiaCheckBalanceParams, HomeDepotCheckBalanceParams
from .crackers.unlocker import EtsyRegisterParams, PrizepicksLoginParams, FifaLoginParams
from .crackers.unlocker import AuspostTrackParams, NorseSearchFlightsParams, ScootBookingParams
from .crackers.unlocker import MyPrepaidValidateParams, PBandaiCartParams, MicrosoftCreateParams
from .crackers.unlocker import GodaddyLoginParams, UltimateDiningLoginParams, FlightInfo, PassengerInfo
from .crackers.unlocker import AjetAvailabilityParams, AttCheckParams, EasyjetFindParams, AjetPnrLoginParams
from .crackers.unlocker import EurowingsBookingParams, FastPeopleSearchParams, HiltonTokenParams, AjetGetBookingDetailParams
from .crackers.unlocker import HiltonCustomerParams, KoreanAirLoginParams, MaerskTokenParams, MarriottLoginParams
from .crackers.unlocker import MarriottEnrollParams, PokemonLoginParams, PokemonLotteryListParams, PokemonApplyLotteryParams
from .crackers.unlocker import SephoraCheckBalanceParams, TextnowLoginParams, WalmartCheckBalanceParams, WalmartCaCheckBalanceParams
from .crackers.unlocker import WalmartWalletParams, FlightSegment, WizzairSearchParams, HmCheckBalanceParams


__all__ = [
    'pynocaptcha', 'magneto',
    'CloudFlareCracker', 'IncapsulaReese84Cracker', 'IncapsulaUtmvcCracker', 'IncapsulaRbzidCracker', 'HcaptchaCracker', 
    'AkamaiV2Cracker', 'ReCaptchaUniversalCracker', 'ReCaptchaEnterpriseCracker', 'ReCaptchaSteamCracker',
    'TlsV1Cracker', 'DiscordCracker', 'ReCaptchaAppCracker', 'AwsUniversalCracker', 'PerimeterxCracker', 'KasadaCdCracker', 
    'crack_akamai_v3', 'async_crack_akamai_v3', 'crack_datadome', 'async_crack_datadome',
    'crack_kasada', 'async_crack_kasada', 'crack_perimeterx', 'async_crack_perimeterx',
    'crack_shape_v1', 'async_crack_shape_v1', 'crack_shape_v2', 'async_crack_shape_v2',
    
    'Unlocker', 'AsyncUnlocker', 'run_unlocker', 'async_run_unlocker',
    'KohlCheckBalanceParams', 'LululemonCheckBalanceParams', 'WbiParams', 'DicksCheckBalanceParams', 
    'FinishlineCheckBalanceParams', 'FootlockerCheckBalanceParams', 'ReiCheckBalanceParams', 
    'One4AllCheckBalanceParams', 'SaksCheckBalanceParams', 'CoachCheckBalanceParams', 
    'FandangoCheckBalanceParams', 'ColumbiaCheckBalanceParams', 'HomeDepotCheckBalanceParams', 
    'EtsyRegisterParams', 'PrizepicksLoginParams', 'FifaLoginParams', 'AuspostTrackParams', 
    'NorseSearchFlightsParams', 'ScootBookingParams', 'MyPrepaidValidateParams', 'PBandaiCartParams', 
    'MicrosoftCreateParams', 'GodaddyLoginParams', 'UltimateDiningLoginParams', 'FlightInfo', 
    'PassengerInfo', 'AjetAvailabilityParams', 'AttCheckParams', 'EasyjetFindParams', 
    'EurowingsBookingParams', 'FastPeopleSearchParams', 'HiltonTokenParams', 'HiltonCustomerParams', 
    'KoreanAirLoginParams', 'MaerskTokenParams', 'MarriottLoginParams', 'MarriottEnrollParams', 
    'PokemonLoginParams', 'PokemonLotteryListParams', 'PokemonApplyLotteryParams', 'AjetPnrLoginParams',
    'SephoraCheckBalanceParams', 'TextnowLoginParams', 'WalmartCheckBalanceParams', 'AjetGetBookingDetailParams',
    'WalmartCaCheckBalanceParams', 'WalmartWalletParams', 'FlightSegment', 'WizzairSearchParams', 
    'HmCheckBalanceParams'
]



# -*- coding: UTF-8 -*-


from .base import BaseCracker


class HcaptchaCracker(BaseCracker):
    
    cracker_name = "hcaptcha"
    cracker_version = "universal"    

    """
    hcaptcha cracker
    :param sitekey: hcaptcha 对接网站的 key
    :param rqtoken: 验证码配置接口有返回 captcha_rqdata、captcha_rqtoken 的请携带该值(如 discord 加频道)
    :param user_agent: 请求流程使用 ua, 必须使用 MacOS Firefox User-Agent, 否则可能破解失败
    :param referer: 触发 hcaptcha 验证的页面地址
    :param mode: 验证模式, 默认 picture 图片验证模式, 可选 question 题库模式
    调用示例:
    cracker = HcaptchaCracker(
        user_token="xxx",
        sitekey='f5561ba9-8f1e-40ca-9b5b-a0b3f719ef34',
        rqdata="",
        referer="https://discord.com/login",

        # debug=True,
    )
    ret = cracker.crack()
    """
    
    # 必传参数
    must_check_params = ["sitekey", "referer"]
    # 默认可选参数
    option_params = {
        "mode": "picture",   # 验证模式, 默认图片验证, 可选 question 题库验证
        "rqdata": "",
        "need_ekey": False,
        "region": None,
        "domain": "hcaptcha.com",
    }


class HcaptchaV2Cracker(BaseCracker):
    
    cracker_name = "hcaptcha"
    cracker_version = "v2"    

    """
    hcaptcha cracker
    :param sitekey: hcaptcha 对接网站的 key
    :param rqtoken: 验证码配置接口有返回 captcha_rqdata、captcha_rqtoken 的请携带该值(如 discord 加频道)
    :param user_agent: 请求流程使用 ua, 必须使用 MacOS Firefox User-Agent, 否则可能破解失败
    :param referer: 触发 hcaptcha 验证的页面地址
    :param mode: 验证模式, 默认 picture 图片验证模式, 可选 question 题库模式
    调用示例:
    cracker = HcaptchaCracker(
        user_token="xxx",
        sitekey='f5561ba9-8f1e-40ca-9b5b-a0b3f719ef34',
        rqdata="",
        referer="https://discord.com/login",

        # debug=True,
    )
    ret = cracker.crack()
    """
    
    # 必传参数
    must_check_params = ["sitekey", "referer"]
    # 默认可选参数
    option_params = {
        "action": None,
        "rqdata": "",
        "need_ekey": False,
        "title": None,
        "extra": {},
        "region": None,
        "ip": None,
        "timezone": None,
        "geolocation": None,
    }



# -*- coding: UTF-8 -*-

import sys
import re
from curl_cffi import requests
from loguru import logger

from typing import Optional, Literal, Union, Tuple, Dict, Any

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from typing import Literal
else:  # pragma: no cover (py38+)
    from typing_extensions import Literal

from .base import BaseCracker
from ..magneto.session import Session, AsyncSession

class AkamaiV2Cracker(BaseCracker):
    
    cracker_name = "akamai"
    cracker_version = "v2"    

    """
    akamai v2 cracker
    :param href: 触发验证的页面地址
    :param api: akamai 提交 sensor_data 的地址
    :param telemetry: 是否 headers 中的 telemetry 参数验证形式, 默认 false
    :param cookies: 请求 href 首页返回的 cookie _abck, bm_sz 值, 传了 api 参数必须传该值, 示例: { "value": "_abck=xxx; bm_sz=xxx", "uri": "https://example.com" }
    :param device: 请求流程使用的设备类型, 可选 pc/mobile, 默认 mobile
    调用示例:
    cracker = AkamaiV2Cracker(
        user_token="xxx",
        href="xxx",
        api="xxx",
        
        # debug=True,
        # proxy=proxy,
    )
    ret = cracker.crack()
    """
    
    # 必传参数
    must_check_params = ["href"]
    # 默认可选参数
    option_params = {
        "branch": "Master",
        "api": "",
        "telemetry": False,
        "uncheck": False,
        "sec_cpt_provider": None,
        "sec_cpt_script": None,
        "sec_cpt_key": None,
        "sec_cpt_challenge": {},
        "sec_cpt_host": None,
        "sec_cpt_html": None,
        "sec_cpt_duration": None,
        "sec_cpt_src": None,
        "sec_cpt_html": None,
        "proxy": None,
        "cookies": {},
        "country": None,
        "ip": None,
        "timezone": None,
        "geolocation": None,
        "user_agent": None,
        "timeout": 30
    }


async def async_crack_akamai_v3(
    user_token: str, requests_args: Dict[str, str], 
    other_requests: Optional[callable] = None, parse_index: Optional[str] = None, accept_language: Optional[str] = None,
    internal_host: bool = True, branch: Optional[str] = None,
    developer_id: Optional[str] = None, auth: bool = False,
    session: Optional[AsyncSession] = None, cookies: Dict[str, str] = {}, proxy: Optional[str] = None,
    timeout: int = 15, allow_redirects: bool = True, verify: bool = False, http_version: Optional[Literal["http1", "http2"]] = None,
    user_agent: Union[str, Tuple[int, Literal["windows", "macos"], Literal["chrome", "edge"]]] = None,
    ipinfo: Dict[str, str] = {}, extra: Dict[str, str] = {},
    debug: bool = False
) -> Tuple[AsyncSession, Dict[str, Any], Dict[str, Any]]:    
    href = requests_args["referer"]
    origin = "/".join(href.split("/")[0:3])
    
    if not session:
        session = AsyncSession(
            user_agent=user_agent, 
            proxy=proxy, 
            ipinfo=ipinfo,
            cookies=cookies, 
            timeout=timeout, 
            allow_redirects=allow_redirects, 
            verify=verify, 
            http_version=http_version
        )
        if not ipinfo:
            await session.update_ipinfo()

    if not extra:
        extra.update(session.client_hints)
        
    user_agent = session.user_agent
    proxy = session.proxy
    
    headers = [
        f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
        'accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7', 
        'upgrade-insecure-requests: 1', 
        "sec-ch-ua-mobile: ?0",
        f"user-agent: {user_agent}",
        f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
        "sec-fetch-site: none",
        "sec-fetch-mode: navigate",
        "sec-fetch-dest: document",
        'sec-fetch-user: ?1',
        "accept-encoding: gzip, deflate, br, zstd",
        f"accept-language: {session.client_hints['accept-language']}",
        "priority: u=0, i"
    ]
    
    response = await session.get(href, headers=headers)
    html = await response.async_text()
    async with requests.AsyncSession() as _session:
        nocaptcha_host = "api.nocaptcha.cn" if internal_host else "api.nocaptcha.io"
        if 'Challenge Validation' in html:
            if debug:
                logger.debug('触发 sec_cpt 验证')
            
            nocaptcha_resp = (await _session.post(
                f'http://{nocaptcha_host}/api/wanda/akamai/v2', headers={
                    'user-token': user_token,
                    **({ "Developer-Id": developer_id } if developer_id else {})
                }, json={
                    'is_auth': auth,
                    'branch': branch,
                    'href': href,
                    'sec_cpt_html': html,
                    'cookies': session.cookies,
                    'user_agent': user_agent,
                    'proxy': session.proxy,
                    **session.ipinfo
                }
            )).json()
            
            if nocaptcha_resp["status"]:
                extra.update(nocaptcha_resp['extra'])
                session.cookies.update(nocaptcha_resp["data"])
                response = await session.get(href, headers=headers)
                html = await response.async_text()
            else:
                raise Warning(f'akamai sec_cpt 验证失败, id: {nocaptcha_resp["id"]}, err: {nocaptcha_resp["msg"]}')
        
        if 'Oops, Something Went Wrong.' in html:
            raise Warning('ip 被 ban, 切 ip 重试')

        if "var chlgeId = ''" in html:
            if debug:
                logger.debug('触发 bm_sc 模式')
            
            bm_sc_src = re.search(r'src="(.*?)"', html)[1].replace('amp;', '')

            nocaptcha_resp = (await _session.post(
                f'http://{nocaptcha_host}/api/wanda/akamai/v2', headers={
                    'user-token': user_token,
                    **({ "Developer-Id": developer_id } if developer_id else {})
                }, json={
                    'is_auth': auth,
                    'branch': branch,
                    'href': href,
                    'bm_sc_src': bm_sc_src,
                    'cookies': session.cookies,
                    'user_agent': user_agent,
                    'proxy': session.proxy,
                    **session.ipinfo
                }
            )).json()
            if debug:
                logger.debug(f"bm_sc 模式验证结果: {nocaptcha_resp}")
            if nocaptcha_resp["status"]:      
                extra.update(nocaptcha_resp['extra'])          
                session.cookies.update(nocaptcha_resp["data"])
                response = await session.get(href, headers=headers)
            else:
                raise Warning(f'akamai bm_sc 验证失败, id: {nocaptcha_resp["id"]}, err: {nocaptcha_resp["msg"]}')
        
        interstitial_retry_times = 0
        while 'triggerInterstitialChallenge' in html and '/_sec/verify?provider=interstitial' in html and interstitial_retry_times < 4:
            logger.debug('触发 bm-verify pow 验证')
            nocaptcha_resp = (await _session.post(
                f'http://{nocaptcha_host}/api/wanda/akamai/v2', headers={
                    'user-token': user_token,
                    **({ "Developer-Id": developer_id } if developer_id else {})
                }, json={
                    'is_auth': auth,
                    'branch': branch,
                    'href': href,
                    'bm_html': html,
                    'cookies': session.cookies,
                    'user_agent': user_agent,
                    'proxy': session.proxy,
                    **session.ipinfo
                }
            )).json()
            if debug:
                logger.debug(nocaptcha_resp)
                
            if nocaptcha_resp["status"]:  
                extra.update(nocaptcha_resp['extra'])
                session.cookies.update(nocaptcha_resp['data'])
                index_resp = session.get(href, headers=headers)
                html = await index_resp.async_text()
            else:
                raise Warning("bm-verify 验证失败")
            
            interstitial_retry_times += 1
        
        if parse_index:
            await parse_index(response, extra)
        
    api = requests_args.get("api")
    if not api:
        try:
            apis = re.findall(r"type=\"text\/javascript\"[ ]{1,2}src\=\"((?:\/[A-Za-z0-9\-\_\+]*?)+)\"(?: defer){0,1}\>\<\/script\>", html)
            if not apis:
                apis = re.findall(r"nonce=\"[a-f0-9]{32}\"[ ]{1,2}src\=\"((?:\/[A-Za-z0-9\-\_\+]*?)+)\"(?: defer){0,1}\>\<\/script\>", html)
            if not apis:
                raise Warning('api 查找失败')
            api = apis[0]
            if not api.startswith("http"):
                api = requests_args.get("api_host", origin) + api
        except:
            raise Warning('api 查找失败')
        
        headers = [
            f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
            f"user-agent: {user_agent}",
            f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
            "sec-ch-ua-mobile: ?0",
            "accept: */*",
            "sec-fetch-site: same-origin",
            "sec-fetch-mode: no-cors",
            "sec-fetch-dest: script",
            f"referer: {href}",
            "accept-encoding: gzip, deflate, br, zstd",
            f"accept-language: {session.client_hints['accept-language']}",
            "priority: u=1"
        ]
        api_response = await session.get(api, headers=headers)
        if api_response.status_code != 200:
            raise Warning("脚本请求失败")

        requests_args['api'] = api
        requests_args["api_headers"] = api_response.headers
    
        iframe_src = requests_args.get("iframe_src")
        if iframe_src:
            nocaptcha_resp = (await _session.post(
                f'http://{nocaptcha_host}/api/wanda/akamai/v2', headers={
                    'user-token': user_token,
                    **({ "Developer-Id": developer_id } if developer_id else {})
                }, json={
                    'is_auth': auth,
                    'branch': branch,
                    'href': href,
                    'api': api,
                    'api_headers': api_response.headers,
                    'cookies': session.cookies,
                    'user_agent': user_agent,
                    'proxy': session.proxy,
                    **session.ipinfo
                }
            )).json()
            
            if nocaptcha_resp['status']:
                extra.update(nocaptcha_resp['extra'])
                session.cookies.update(nocaptcha_resp["data"])                        

                headers = [
                    f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
                    "sec-ch-ua-mobile: ?0",
                    f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
                    'upgrade-insecure-requests: 1',                 
                    f"user-agent: {user_agent}",
                    'accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7', 
                    "sec-fetch-site: same-origin",
                    "sec-fetch-mode: navigate",
                    'sec-fetch-user: ?1',
                    "sec-fetch-dest: iframe",
                    f'referer: {href}',
                    "accept-encoding: gzip, deflate, br, zstd",
                    f"accept-language: {session.client_hints['accept-language']}",
                    "priority: u=0, i"
                ]
                response = await session.get(iframe_src, headers=headers)                    
                html = await response.async_text()
                
                origin = "/".join(iframe_src.split("/")[0:3])
                try:
                    apis = re.findall(r"type=\"text\/javascript\"[ ]{1,2}src\=\"((?:\/[A-Za-z0-9\-\_\+]*?)+)\"(?: defer){0,1}\>\<\/script\>", html)
                    if not apis:
                        apis = re.findall(r"nonce=\"[a-f0-9]{32}\"[ ]{1,2}src\=\"((?:\/[A-Za-z0-9\-\_\+]*?)+)\"(?: defer){0,1}\>\<\/script\>", html)
                    if not apis:
                        raise Warning('api 查找失败')
                    api = apis[0]
                    if not api.startswith("http"):
                        api = requests_args.get("api_host", origin) + api
                except:
                    raise Warning('api 查找失败')

                headers = [
                    f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
                    f"user-agent: {user_agent}",
                    f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
                    "sec-ch-ua-mobile: ?0",
                    "accept: */*",
                    "sec-fetch-site: same-origin",
                    "sec-fetch-mode: no-cors",
                    "sec-fetch-dest: script",
                    f"referer: {iframe_src}",
                    "accept-encoding: gzip, deflate, br, zstd",
                    f"accept-language: {session.client_hints['accept-language']}",
                    "priority: u=2"
                ]

                api_response = await session.get(api, headers=headers)
                if api_response.status_code != 200:
                    raise Warning(f"iframe api 脚本请求失败: {api_response.status_code}")

                requests_args["referer"] = iframe_src
                requests_args['api'] = api
                requests_args["api_headers"] = api_response.headers
                del requests_args["iframe_src"]
            else:
                raise Warning("akamai v2 验证失败")
        
        if other_requests:
            await other_requests(session, requests_args, extra)

        for k, v in requests_args.items():
            if isinstance(v, dict):
                for k1, v1 in v.items():
                    if callable(v1):
                        requests_args[k][k1] = v1(response)
        
        sub_domain = '.' + '.'.join(href.split('/')[2].split('.')[-2:])
        
        nocaptcha_resp = (await _session.post(
            f'http://{nocaptcha_host}/api/wanda/akamai/v3', headers={
                'user-token': user_token,
                **({ "Developer-Id": developer_id } if developer_id else {})
            }, json={
                'is_auth': auth,
                'branch': branch,
                **requests_args,
                'cookies': [
                    {
                        'name': name,
                        'value': value,
                        'domain': sub_domain
                    } for name, value in session.cookies.items()
                ],
                'user_agent': user_agent,
                'proxy': session.proxy,
                **session.ipinfo
            }
        )).json()
        if debug:
            logger.debug(f"验证结果: {nocaptcha_resp}")
            
        if nocaptcha_resp["status"]:
            extra.update(nocaptcha_resp['extra'])
            session.cookies.update(nocaptcha_resp["data"]["cookies"])
            return session, nocaptcha_resp["data"].get("response", {}), extra
        else:
            raise Warning(f'验证失败, id: {nocaptcha_resp["id"]}, err: {nocaptcha_resp["msg"]}')


def crack_akamai_v3(
    user_token: str, requests_args: Dict[str, str],
    other_requests: Optional[callable] = None, parse_index: Optional[callable] = None,
    internal_host: bool = True, branch: Optional[str] = None,
    developer_id: Optional[str] = None, auth: bool = False,
    session: Optional[Session] = None, cookies: Dict[str, str] = {}, proxy: Optional[str] = None,
    timeout: int = 15, allow_redirects: bool = True, verify: bool = False, http_version: Optional[Literal["http1", "http2"]] = None,
    user_agent: Union[str, Tuple[int, Literal["windows", "macos"], Literal["chrome", "edge"]]] = None,
    ipinfo: Dict[str, str] = {}, extra: Dict[str, str] = {},
    debug: bool = False
) -> Tuple[Session, Dict[str, Any], Dict[str, Any]]:
    href = requests_args["referer"]
    origin = "/".join(href.split("/")[0:3])
    
    if not session:
        session = Session(
            user_agent=user_agent, 
            proxy=proxy, 
            ipinfo=ipinfo,
            cookies=cookies, 
            timeout=timeout, 
            allow_redirects=allow_redirects, 
            verify=verify, 
            http_version=http_version
        )
        if not ipinfo:
            session.update_ipinfo()

    if not extra:
        extra.update(session.client_hints)
        
    user_agent = session.user_agent
    proxy = session.proxy
    
    headers = [
        f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
        'accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7', 
        'upgrade-insecure-requests: 1', 
        "sec-ch-ua-mobile: ?0",
        f"user-agent: {user_agent}",
        f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
        "sec-fetch-site: none",
        "sec-fetch-mode: navigate",
        "sec-fetch-dest: document",
        'sec-fetch-user: ?1',
        "accept-encoding: gzip, deflate, br, zstd",
        f"accept-language: {session.client_hints['accept-language']}",
        "priority: u=0, i"
    ]
    
    response = session.get(href, headers=headers)

    nocaptcha_host = "api.nocaptcha.cn" if internal_host else "api.nocaptcha.io"
    
    html = response.text
    if 'Challenge Validation' in html:
        if debug:
            logger.debug('触发 sec_cpt 验证')
        
        nocaptcha_resp = requests.post(
            f'http://{nocaptcha_host}/api/wanda/akamai/v2', headers={
                'user-token': user_token,
                **({ "Developer-Id": developer_id } if developer_id else {})
            }, json={
                'is_auth': auth,
                'branch': branch,
                'href': href,
                'sec_cpt_html': html,
                'cookies': session.cookies,
                'user_agent': user_agent,
                'proxy': session.proxy,
                **session.ipinfo
            }
        ).json()
        if debug:
            logger.debug(nocaptcha_resp)
        
        if nocaptcha_resp["status"]:
            extra.update(nocaptcha_resp['extra'])
            session.cookies.update(nocaptcha_resp["data"])
            response = session.get(href, headers=headers)
        else:
            raise Warning(f'akamai sec_cpt 验证失败, id: {nocaptcha_resp["id"]}, err: {nocaptcha_resp["msg"]}')
    
    if 'Oops, Something Went Wrong.' in html:
        raise Warning('ip 被 ban, 切 ip 重试')
    
    if "var chlgeId = ''" in html:
        if debug:
            logger.debug('触发 bm_sc 模式')
        
        bm_sc_src = re.search(r'src="(.*?)"', html)[1].replace('amp;', '')

        nocaptcha_resp = requests.post(
            f'http://{nocaptcha_host}/api/wanda/akamai/v2', headers={
                'user-token': user_token,
                **({ "Developer-Id": developer_id } if developer_id else {})
            }, json={
                'is_auth': auth,
                'branch': branch,
                'href': href,
                'bm_sc_src': bm_sc_src,
                'cookies': session.cookies,
                'user_agent': user_agent,
                'proxy': session.proxy,
                **session.ipinfo
            }
        ).json()
        if debug:
            logger.debug(nocaptcha_resp)
            
        if nocaptcha_resp["status"]:                
            extra.update(nocaptcha_resp['extra'])
            session.cookies.update(nocaptcha_resp["data"])
            response = session.get(href, headers=headers)
        else:
            raise Warning(f'akamai bm_sc 验证失败, id: {nocaptcha_resp["id"]}, err: {nocaptcha_resp["msg"]}')

    interstitial_retry_times = 0
    while 'triggerInterstitialChallenge' in html and '/_sec/verify?provider=interstitial' in html and interstitial_retry_times < 4:
        logger.debug('触发 bm-verify pow 验证')
        nocaptcha_resp = requests.post(
            f'http://{nocaptcha_host}/api/wanda/akamai/v2', headers={
                'user-token': user_token,
                **({ "Developer-Id": developer_id } if developer_id else {})
            }, json={
                'is_auth': auth,
                'branch': branch,
                'href': href,
                'bm_html': html,
                'cookies': session.cookies,
                'user_agent': user_agent,
                'proxy': session.proxy,
                **session.ipinfo
            }
        ).json()
        if debug:
            logger.debug(nocaptcha_resp)
            
        if nocaptcha_resp["status"]:  
            extra.update(nocaptcha_resp['extra'])
            session.cookies.update(nocaptcha_resp['data'])
            index_resp = session.get(href, headers=headers)
            html = index_resp.text
        else:
            raise Warning("bm-verify 验证失败")
        
        interstitial_retry_times += 1
        
    if parse_index:
        parse_index(response, extra)
    
    api = requests_args.get("api")
    if not api:
        try:
            apis = re.findall(r"type=\"text\/javascript\"[ ]{1,2}src\=\"((?:\/[A-Za-z0-9\-\_\+]*?)+)\"(?: defer){0,1}\>\<\/script\>", html)
            if not apis:
                apis = re.findall(r"nonce=\"[a-f0-9]{32}\"[ ]{1,2}src\=\"((?:\/[A-Za-z0-9\-\_\+]*?)+)\"(?: defer){0,1}\>\<\/script\>", html)
            if not apis:
                raise Warning('api 查找失败')
            api = apis[0]
            if not api.startswith("http"):
                api = requests_args.get("api_host", origin) + api
        except:
            raise Warning('api 查找失败')
    
    headers = [
        f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
        f"user-agent: {user_agent}",
        f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
        "sec-ch-ua-mobile: ?0",
        "accept: */*",
        "sec-fetch-site: same-origin",
        "sec-fetch-mode: no-cors",
        "sec-fetch-dest: script",
        f"referer: {href}",
        "accept-encoding: gzip, deflate, br, zstd",
        f"accept-language: {session.client_hints['accept-language']}",
        "priority: u=1"
    ]
    api_response = session.get(api, headers=headers)
    if api_response.status_code != 200:
        raise Warning("脚本请求失败")

    requests_args['api'] = api
    requests_args["api_headers"] = api_response.headers

    iframe_src = requests_args.get("iframe_src")
    if iframe_src:
        nocaptcha_resp = requests.post(
            f'http://{nocaptcha_host}/api/wanda/akamai/v2', headers={
                'user-token': user_token,
                **({ "Developer-Id": developer_id } if developer_id else {})
            }, json={
                'is_auth': auth,
                'branch': branch,
                'href': href,
                'api': api,
                'api_headers': api_response.headers,
                'cookies': session.cookies,
                'user_agent': user_agent,
                'proxy': session.proxy,
                **session.ipinfo
            }
        ).json()
        if debug:
            logger.debug(nocaptcha_resp)
        
        if nocaptcha_resp['status']:
            extra.update(nocaptcha_resp['extra'])
            session.cookies.update(nocaptcha_resp["data"])                        

            headers = [
                f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
                "sec-ch-ua-mobile: ?0",
                f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
                'upgrade-insecure-requests: 1',                 
                f"user-agent: {user_agent}",
                'accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7', 
                "sec-fetch-site: same-origin",
                "sec-fetch-mode: navigate",
                'sec-fetch-user: ?1',
                "sec-fetch-dest: iframe",
                f'referer: {href}',
                "accept-encoding: gzip, deflate, br, zstd",
                f"accept-language: {session.client_hints['accept-language']}",
                "priority: u=0, i"
            ]
            response = session.get(iframe_src, headers=headers)
            
            origin = "/".join(iframe_src.split("/")[0:3])
            html = response.text
            try:
                apis = re.findall(r"type=\"text\/javascript\"[ ]{1,2}src\=\"((?:\/[A-Za-z0-9\-\_\+]*?)+)\"(?: defer){0,1}\>\<\/script\>", html)
                if not apis:
                    apis = re.findall(r"nonce=\"[a-f0-9]{32}\"[ ]{1,2}src\=\"((?:\/[A-Za-z0-9\-\_\+]*?)+)\"(?: defer){0,1}\>\<\/script\>", html)
                if not apis:
                    raise Warning('api 查找失败')
                api = apis[0]
                if not api.startswith("http"):
                    api = requests_args.get("api_host", origin) + api
            except:
                raise Warning('api 查找失败')

            headers = [
                f"sec-ch-ua-platform: {session.client_hints['sec-ch-ua-platform']}",
                f"user-agent: {user_agent}",
                f"sec-ch-ua: {session.client_hints['sec-ch-ua']}",
                "sec-ch-ua-mobile: ?0",
                "accept: */*",
                "sec-fetch-site: same-origin",
                "sec-fetch-mode: no-cors",
                "sec-fetch-dest: script",
                f"referer: {iframe_src}",
                "accept-encoding: gzip, deflate, br, zstd",
                f"accept-language: {session.client_hints['accept-language']}",
                "priority: u=2"
            ]

            api_response = session.get(api, headers=headers)
            if api_response.status_code != 200:
                raise Warning(f"iframe api 脚本请求失败: {api_response.status_code}")

            requests_args["referer"] = iframe_src
            requests_args['api'] = api
            requests_args["api_headers"] = api_response.headers
            del requests_args["iframe_src"]
        else:
            raise Warning("akamai v2 验证失败")
    
    if other_requests:
        other_requests(session, requests_args, extra)

    for k, v in requests_args.items():
        if isinstance(v, dict):
            for k1, v1 in v.items():
                if callable(v1):
                    requests_args[k][k1] = v1(response)
    
    sub_domain = '.' + '.'.join(href.split('/')[2].split('.')[-2:])
    
    nocaptcha_resp = requests.post(
        f'http://{nocaptcha_host}/api/wanda/akamai/v3', headers={
            'user-token': user_token,
            **({ "Developer-Id": developer_id } if developer_id else {})
        }, json={
            'is_auth': auth,
            'branch': branch,
            **requests_args,
            'cookies': [
                {
                    'name': name,
                    'value': value,
                    'domain': sub_domain
                } for name, value in session.cookies.items()
            ],
            'user_agent': user_agent,
            'proxy': session.proxy,
            **session.ipinfo
        }
    ).json()
    if debug:
        logger.debug(nocaptcha_resp)
        
    if nocaptcha_resp["status"]:
        extra.update(nocaptcha_resp['extra'])
        session.cookies.update(nocaptcha_resp["data"]["cookies"])
        return session, nocaptcha_resp["data"].get("response", {}), extra
    else:
        raise Warning(f'验证失败, id: {nocaptcha_resp["id"]}, err: {nocaptcha_resp["msg"]}')



# -*- coding: utf-8 -*-

from typing import Dict, Any, Union

import rnet
from json import loads


class Response():
    
    def __init__(self, response: Union[rnet.BlockingResponse, rnet.Response]):
        self._response = response
        self._text: str = None
        self._headers: Dict[str, str] = None
        self._async = isinstance(response, rnet.Response)

    def __getattr__(self, name):
        """
        当访问 Response 没有定义的属性或方法时，尝试从 self._response 获取。
        这样可以支持 self._response 的所有方法和属性。
        """
        # 获取 self._response 的属性或方法
        attr = getattr(self._response, name)

        # 如果是方法，返回一个包装器以确保调用时正确传递参数
        if callable(attr):
            def wrapper(*args, **kwargs):
                return attr(*args, **kwargs)
            return wrapper
        else:
            return attr
        
    @property
    def status_code(self) -> int:
        return self._response.status_code.as_int()
        
    @property
    def cookies(self) -> Dict[str, str]:
        return {
            cookie.name: cookie.value for cookie in self._response.cookies
        }
    
    @property
    def headers(self) -> Dict[str, str]:
        if self._headers is None:
            self._headers = {}
            for header_key, header_value in self._response.headers.items():
                self._headers[header_key.decode()] = header_value.decode()
        return self._headers
    
    @property
    def text(self) -> str:
        if self._text is None:
            if self._async:
                raise Exception("Using async_text instead")
            else:
                self._text = self._response.text()
                self._response.close()
        return self._text

    async def async_text(self) -> str:
        if self._text is None:
            if not self._async:
                self._text = self._response.text()
                self._response.close()
            else:
                try:
                    self._text = await self._response.text()
                finally:
                    await self._response.close()
        return self._text

    def json(self) -> Dict[str, Any]:
        return loads(self.text)

    async def async_json(self) -> Dict[str, Any]:
        await self.async_text()
        return self.json()
